"""
ABSFUYU CLI
-----------
Game

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["game_group"]


# Library
# ---------------------------------------------------------------------------
import click

from absfuyu.game import game_escapeLoop, game_RockPaperScissors
from absfuyu.game.schulte import SchulteTable
from absfuyu.game.sudoku import Sudoku
from absfuyu.game.tictactoe import GameMode, TicTacToe
from absfuyu.game.wordle import Wordle


# CLI
# ---------------------------------------------------------------------------
@click.command
@click.option(
    "--hard",
    "hard_mode",
    is_flag=True,
    default=False,
    show_default=True,
    help="Hard mode",
)
def rps(hard_mode: bool) -> None:
    """Game: Rock Paper Scissors"""
    game_RockPaperScissors(hard_mode=hard_mode)


@click.command(name="esl")
def escape_loop() -> None:
    """Game: Escape loop"""
    game_escapeLoop()


@click.command(name="wordle")
def wordle_solver() -> None:
    """Worldle solver"""
    Wordle().solve()


@click.command(name="sudoku")
@click.argument("sudoku_string", type=str)
@click.option(
    "--raw",
    "raw",
    is_flag=True,
    default=False,
    show_default=True,
    help="Returns answer in string form",
)
def sudoku_solver(sudoku_string: str, raw: bool) -> None:
    """Sudoku solver"""
    sdk = Sudoku.from_string(sudoku_string)
    if raw:
        click.echo(sdk.solve().export_to_string())
    else:
        click.echo(sdk.solve().to_board_form())


@click.command(name="ttt")
@click.option(
    "--size",
    "-s",
    "size",
    type=int,
    default=3,
    show_default=True,
    help="Size of the board",
)
@click.option(
    "--mode",
    "-m",
    "game_mode",
    type=click.Choice([GameMode.ONE_V_BOT, GameMode.ONE_V_ONE, GameMode.BOT_V_BOT]),
    default=GameMode.ONE_V_BOT,
    show_default=True,
    help="Game mode",
)
@click.option(
    "--bot-time",
    "-t",
    "bot_time",
    type=float,
    default=0.0,
    show_default=True,
    help="Time between bot move",
)
def tictactoe(size: int, game_mode: str, bot_time: float) -> None:
    """Game: Tictactoe"""
    instance = TicTacToe(size)
    instance.play(game_mode, bot_time=bot_time)


@click.command(name="schulte")
@click.option(
    "--size",
    "-s",
    "size",
    type=int,
    default=5,
    show_default=True,
    help="Size of the table",
)
def schulte_table(size: int) -> None:
    """Schulte table"""
    engine = SchulteTable(size=size)
    engine.make_table()


@click.group(name="game")
def game_group() -> None:
    """Play game"""
    pass


game_group.add_command(rps)
game_group.add_command(escape_loop)
game_group.add_command(wordle_solver)
game_group.add_command(sudoku_solver)
game_group.add_command(tictactoe)
game_group.add_command(schulte_table)



"""
Absfuyu: Shape
--------------
Shapes

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    # Polygon
    "Triangle",
    "Circle",
    "Square",
    "Rectangle",
    "Pentagon",
    "Hexagon",
    "Parallelogram",
    "Rhombus",
    "Trapezoid",
    # 3D Shape
    "Cube",
    "Cuboid",
    "Sphere",
    "HemiSphere",
    "Cylinder",
]


# Library
# ---------------------------------------------------------------------------
import math
from abc import ABC, abstractmethod
from typing import ClassVar, Self

try:
    from typing import override  # type: ignore
except ImportError:
    from absfuyu.core.decorator import dummy_decorator as override

from absfuyu.core import BaseClass


# Class
# ---------------------------------------------------------------------------
class Shape(BaseClass):
    """Shape base class"""

    pass


class Polygon(ABC, Shape):
    """2D Shape class base"""

    POLYGON_LIST: ClassVar[list[str]] = []

    def __init__(self, num_of_sides: int) -> None:
        """
        Initialize a polygon with number of sides.

        Parameters
        ----------
        num_of_sides : int
            Number of sides of the polygon.

        Raises
        ------
        ValueError
            If the number of sides smaller than 3.
        """
        if num_of_sides <= 2:
            raise ValueError("Number of sides must be larger than 2.")

        self._num_of_sides = num_of_sides

    def __init_subclass__(cls, *args, **kwargs) -> None:
        super().__init_subclass__(*args, **kwargs)
        # cls.POLYGON_LIST.append(cls)  # append type
        cls.POLYGON_LIST.append(cls.__name__)  # append name

    @abstractmethod
    def perimeter(self):
        pass

    @abstractmethod
    def area(self):
        pass


class EqualSidesPolygon(Polygon):
    """
    Base class for polygons with equal side length
    """

    def __init__(self, side: int | float, num_of_sides: int) -> None:
        """
        Initialize a polygon with equal side length.

        Parameters
        ----------
        side : int | float
            Length of each side.

        num_of_sides : int
            Number of sides of the polygon.

        Raises
        ------
        ValueError
            If the side is not positive or number of sides smaller than 3.
        """
        super().__init__(num_of_sides=num_of_sides)

        if side <= 0:
            raise ValueError("Side length must be a positive number.")

        self.side = side

    @override
    def perimeter(self) -> int | float:
        """
        Calculate the perimeter of the polygon.

        Returns
        -------
        int | float
            The perimeter of the polygon.
        """
        return self.side * self._num_of_sides

    @abstractmethod
    @override
    def area(self):
        pass

    def interior_angle(self) -> float:
        """
        Calculate the interior angle of the polygon.

        Returns
        -------
        float
            The interior angle in degrees.
        """
        return (self._num_of_sides - 2) * 180 / self._num_of_sides


class ThreeDimensionShape(ABC, Shape):
    SHAPE_LIST: ClassVar[list[str]] = []

    def __init_subclass__(cls, *args, **kwargs) -> None:
        super().__init_subclass__(*args, **kwargs)
        # cls.SHAPE_LIST.append(cls)  # append type
        cls.SHAPE_LIST.append(cls.__name__)  # append name

    @abstractmethod
    def surface_area(self):
        pass

    @abstractmethod
    def volume(self):
        pass


# Class - Polygon
# ---------------------------------------------------------------------------
class Triangle(Polygon):
    def __init__(
        self,
        a: int | float,
        b: int | float,
        c: int | float,
    ) -> None:
        """
        Initializes a Triangle instance with three sides.

        Parameters
        ----------
        a : int | float
            The length of the first side.

        b : int | float
            The length of the second side.

        c : int | float
            The length of the third side.

        Raises
        ------
        ValueError
            If any side length is not positive or if the sides do not form a valid triangle.
        """
        super().__init__(num_of_sides=3)

        if a <= 0 or b <= 0 or c <= 0:
            raise ValueError("Side lengths must be positive.")

        # Check for triangle inequality theorem
        if (a + b <= c) or (a + c <= b) or (b + c <= a):
            raise ValueError("The provided lengths do not form a valid triangle.")

        self.a = a
        self.b = b
        self.c = c

    @override
    def perimeter(self) -> int | float:
        """
        Calculates and returns the perimeter of the triangle.
        """
        return self.a + self.b + self.c

    @override
    def area(self) -> int | float:
        """
        Calculates and returns the area of the triangle using Heron's formula.
        """
        s = self.perimeter() / 2  # Semi-perimeter
        # Heron formula
        res = math.sqrt(s * (s - self.a) * (s - self.b) * (s - self.c))
        return res

    def is_right_angled(self) -> bool:
        """
        Checks if the triangle is a right-angled triangle
        (one vertex has degree of 90) using the Pythagorean theorem.

        Returns
        -------
        bool
            ``True`` if the triangle is right-angled, ``False`` otherwise.
        """
        sides = sorted([self.a, self.b, self.c])
        # return (
        #     sides[0] ** 2 + sides[1] ** 2 == sides[2] ** 2
        # )  # Pytagorean formula

        # Using ``==`` to compare floating-point numbers can be
        # unreliable due to precision issues. Use a tolerance instead
        return abs(sides[2] ** 2 - (sides[0] ** 2 + sides[1] ** 2)) < 1e-6

    def is_equilateral(self) -> bool:
        """
        Checks if the triangle is an equilateral triangle
        (3 sides have the same length).

        Returns
        -------
        bool
            ``True`` if the triangle is equilateral, ``False`` otherwise.
        """
        return self.a == self.b and self.b == self.c

    def is_isosceles(self) -> bool:
        """
        Checks if the triangle is an isosceles triangle
        (at least two sides are equal).

        Returns
        -------
        bool
            ``True`` if the triangle is isosceles, ``False`` otherwise.
        """
        return self.a == self.b or self.b == self.c or self.c == self.a

    def triangle_type(self) -> str:
        """
        Determines the type of triangle based on its sides.

        Returns
        -------
        str
            A string describing the type of triangle: ``"equilateral"``, ``"isosceles"``,
            ``"right-angled"``, or ``"scalene"`` if none of the other types apply.
        """

        if self.is_equilateral():
            return "equilateral"
        elif self.is_isosceles():
            if self.is_right_angled():
                return "right-angled isosceles"
            return "isosceles"
        elif self.is_right_angled():
            return "right-angled"
        else:
            return "scalene"

    def is_acute(self) -> bool:
        """
        Checks if the triangle is an acute triangle
        (all angles less than 90 degrees).

        Returns
        -------
        bool
            ``True`` if the triangle is acute, ``False`` otherwise.
        """
        sides = sorted([self.a, self.b, self.c])
        return sides[0] ** 2 + sides[1] ** 2 > sides[2] ** 2

    def is_obtuse(self) -> bool:
        """
        Checks if the triangle is an obtuse triangle
        (one angle greater than 90 degrees).

        Returns
        -------
        bool
            ``True`` if the triangle is obtuse, ``False`` otherwise.
        """
        sides = sorted([self.a, self.b, self.c])
        return sides[0] ** 2 + sides[1] ** 2 < sides[2] ** 2

    def get_angles(self) -> tuple[float, float, float]:
        """
        Calculates and returns the angles of the triangle in degrees.

        Returns
        -------
        tuple[float, float, float]
            A tuple containing the angles in degrees (angle_A, angle_B, angle_C).
        """
        a, b, c = self.a, self.b, self.c
        angle_A = math.degrees(math.acos((b**2 + c**2 - a**2) / (2 * b * c)))
        angle_B = math.degrees(math.acos((a**2 + c**2 - b**2) / (2 * a * c)))
        angle_C = math.degrees(math.acos((a**2 + b**2 - c**2) / (2 * a * b)))
        return angle_A, angle_B, angle_C

    def scale(self, factor: int | float) -> None:
        """
        Scales the triangle by a given factor, changing the lengths of all sides.

        Parameters
        ----------
        factor : int | float
            The scaling factor.  Must be positive.

        Raises
        ------
        ValueError
            If the scaling factor is not positive.
        """
        if factor <= 0:
            raise ValueError("Scaling factor must be positive.")
        self.a *= factor
        self.b *= factor
        self.c *= factor

    def is_similar(self, other: Self) -> bool:
        """
        Checks if this triangle is similar to another triangle.

        Parameters
        ----------
        other : Triangle
            The other triangle to compare to.

        Returns
        -------
        bool
            ``True`` if the triangles are similar, ``False`` otherwise.
        """
        ratios = sorted([self.a / other.a, self.b / other.b, self.c / other.c])
        return abs(ratios[0] - ratios[2]) < 1e-6

    def vis(self) -> str:
        """Visualization of Triangle"""
        out = """
              A
             /\\
         c /....\\b
        B/........\\C
               a
        """
        return out


class Circle(Polygon):
    def __init__(self, radius: int | float) -> None:
        """
        Initializes a Circle instance with a specified radius.

        Parameters
        ----------
        radius : int | float
            The radius of the circle. Must be positive.

        Raises
        ------
        ValueError
            If the radius is not positive.
        """
        if radius <= 0:
            raise ValueError("Radius must be a positive number.")
        self.radius = radius

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Circle):
            raise NotImplementedError("Not supported")
        return math.isclose(self.radius, other.radius)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Circle):
            raise NotImplementedError("Not supported")
        return self.radius < other.radius

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Circle):
            raise NotImplementedError("Not supported")
        return self.radius <= other.radius

    def diameter(self) -> int | float:
        """Returns the diameter of the circle."""
        return 2 * self.radius

    @override
    def perimeter(self) -> float:
        """Returns the circumference of the circle."""
        return 2 * math.pi * self.radius

    def circumference(self) -> float:
        """Returns the circumference of the circle."""
        return self.perimeter()  # type: ignore

    @override
    def area(self) -> float:
        """Returns the area of the circle."""
        return math.pi * self.radius**2

    def scale(self, factor: int | float) -> None:
        """
        Scales the circle by a given factor, changing its radius.

        Parameters
        ----------
        factor : int | float
            The scaling factor. Must be positive.

        Raises
        ------
        ValueError
            If the scaling factor is not positive.
        """
        if factor <= 0:
            raise ValueError("Scaling factor must be positive.")

        self.radius *= factor

    @classmethod
    def from_area(cls, area: int | float) -> Self:
        """
        Creates a Circle instance from its area.

        Parameters
        ----------
        area : int | float
            The area of the circle. Must be positive.

        Raises
        ------
        ValueError
            If the area is not positive.
        """
        if area <= 0:
            raise ValueError("Area must be positive.")

        radius = math.sqrt(area / math.pi)
        return cls(radius)

    @classmethod
    def from_circumference(cls, circumference: int | float) -> Self:
        """
        Creates a Circle instance from its circumference.

        Parameters
        ----------
        circumference : int | float
            The circumference of the circle. Must be positive.

        Raises
        ------
        ValueError
            If the circumference is not positive.
        """
        if circumference <= 0:
            raise ValueError("Circumference must be positive.")

        radius = circumference / (2 * math.pi)
        return cls(radius)


class Square(EqualSidesPolygon):
    def __init__(self, side: int | float) -> None:
        """
        Initializes a Square instance with specified side.

        Parameters
        ----------
        side : int | float
            The length of the rectangle. Must be positive.

        Raises
        ------
        ValueError
            If the side is not positive.
        """
        super().__init__(side=side, num_of_sides=4)

    @override
    def area(self) -> int | float:
        """Calculate the area of the square."""
        return self.side**2

    def diagonal(self) -> float:
        """Calculate the length of the diagonal of the square."""
        return math.sqrt(2) * self.side

    @classmethod
    def from_perimeter(cls, perimeter: int | float) -> Self:
        """Create a Square instance from its perimeter."""
        if perimeter <= 0:
            raise ValueError("Perimeter must be a positive number.")
        return cls(perimeter / 4)

    @classmethod
    def from_area(cls, area: int | float) -> Self:
        """Create a Square instance from its area."""
        if area <= 0:
            raise ValueError("Area must be a positive number.")
        return cls(math.sqrt(area))

    def scale(self, factor: int | float) -> None:
        """
        Scales the square by a given factor.

        Parameters
        ----------
        factor : int | float
            The scaling factor. Must be positive.

        Raises
        ------
        ValueError
            If the scaling factor is not positive.
        """
        if factor <= 0:
            raise ValueError("Scaling factor must be positive.")

        self.side *= factor


class Rectangle(Polygon):
    def __init__(self, length: int | float, width: int | float) -> None:
        """
        Initializes a Rectangle instance with specified length and width.

        Parameters
        ----------
        length : int | float
            The length of the rectangle. Must be positive.
        width : int | float
            The width of the rectangle. Must be positive.

        Raises
        ------
        ValueError
            If either length or width is not positive.
        """
        super().__init__(num_of_sides=4)

        if length <= 0 or width <= 0:
            raise ValueError("Length and width must be positive numbers.")

        self.length = max(length, width)
        self.width = min(length, width)

    @override
    def perimeter(self) -> int | float:
        """Calculates and returns the perimeter of the rectangle."""
        return 2 * (self.length + self.width)

    @override
    def area(self) -> int | float:
        """Calculates and returns the area of the rectangle."""
        return self.length * self.width

    def diagonal(self) -> float:
        """Calculates and returns the length of the diagonal of the rectangle."""
        return math.sqrt(self.length**2 + self.width**2)

    def is_square(self) -> bool:
        """Checks if the rectangle is a square (length equals width)."""
        return self.length == self.width

    def scale(self, factor: int | float) -> None:
        """
        Scales the rectangle by a given factor.

        Parameters
        ----------
        factor : int | float
            The scaling factor. Must be positive.

        Raises
        ------
        ValueError
            If the scaling factor is not positive.
        """
        if factor <= 0:
            raise ValueError("Scaling factor must be positive.")

        self.length *= factor
        self.width *= factor


class Pentagon(EqualSidesPolygon):
    def __init__(self, side: int | float) -> None:
        """
        Initializes a Pentagon instance with a specified side length.

        Parameters
        ----------
        side : int | float
            The length of one side of the pentagon. Must be positive.

        Raises
        ------
        ValueError
            If the side length is not positive.
        """
        super().__init__(side=side, num_of_sides=5)

    @override
    def area(self) -> float:
        """Calculates and returns the area of the pentagon using a specific formula."""
        res = 0.25 * math.sqrt(5 * (5 + 2 * math.sqrt(5))) * self.side**2
        return res

    def apothem(self) -> float:
        """Calculates and returns the apothem of the pentagon."""
        res = (self.side / 2) / math.tan(36 * math.pi / 180)
        return res

    def area2(self) -> float:
        """Calculates the area using the apothem and perimeter."""
        res = 0.5 * self.perimeter() * self.apothem()
        return res  # type: ignore

    @classmethod
    def from_area(cls, area: int | float) -> Self:
        """
        Creates a Pentagon instance from its area.

        Parameters
        ----------
        area : int | float
            The area of the pentagon. Must be positive.

        Raises
        ------
        ValueError
            If the area is not positive.
        """
        if area <= 0:
            raise ValueError("Area must be positive.")

        # Calculate side length from area
        side = math.sqrt((4 * area)) / math.sqrt(math.sqrt(5 * (5 + 2 * math.sqrt(5))))
        return cls(side)

    @classmethod
    def from_perimeter(cls, perimeter: int | float) -> Self:
        """
        Creates a Pentagon instance from its perimeter.

        Parameters
        ----------
        perimeter : int | float
            The perimeter of the pentagon. Must be positive.

        Raises
        ------
        ValueError
            If the perimeter is not positive.
        """
        if perimeter <= 0:
            raise ValueError("Perimeter must be positive.")

        # Calculate side length from perimeter
        side = perimeter / 5
        return cls(side)


class Hexagon(EqualSidesPolygon):
    def __init__(self, side: int | float) -> None:
        """
        Initializes a Hexagon instance with a specified side length.

        Parameters
        ----------
        side : int | float
            The length of one side of the hexagon. Must be positive.

        Raises
        ------
        ValueError
            If the side length is not positive.
        """
        super().__init__(side=side, num_of_sides=6)

    @override
    def area(self) -> float:
        """Calculates and returns the area of the hexagon."""
        res = self.side**2 * (3 * math.sqrt(3) / 2)
        return res

    def apothem(self) -> float:
        """Calculates and returns the apothem of the hexagon."""
        return self.side / (2 * math.tan(math.pi / 6))

    def area2(self) -> float:
        """Calculates the area using the apothem and perimeter."""
        return 0.5 * self.perimeter() * self.apothem()  # type: ignore

    @classmethod
    def from_area(cls, area: int | float) -> Self:
        """
        Creates a Hexagon instance from its area.

        Parameters
        ----------
        area : int | float
            The area of the hexagon. Must be positive.

        Raises
        ------
        ValueError
            If the area is not positive.
        """
        if area <= 0:
            raise ValueError("Area must be positive.")

        # Calculate side length from area
        side = math.sqrt((2 * area) / (3 * math.sqrt(3)))
        return cls(side)

    @classmethod
    def from_perimeter(cls, perimeter: int | float) -> Self:
        """
        Creates a Hexagon instance from its perimeter.

        Parameters
        ----------
        perimeter : int | float
            The perimeter of the hexagon. Must be positive.

        Raises
        ------
        ValueError
            If the perimeter is not positive.
        """
        if perimeter <= 0:
            raise ValueError("Perimeter must be positive.")

        # Calculate side length from perimeter
        side = perimeter / 6
        return cls(side)


class Parallelogram(Polygon):
    def __init__(
        self,
        base: int | float,
        height: int | float,
        *,
        a: int | float | None = None,
        phi: int | float | None = None,
    ) -> None:
        """
        Initializes a Parallelogram instance with specified dimensions.

        Parameters
        ----------
        base : int | float
            The length of the base of the parallelogram. Must be positive.

        height : int | float
            The height of the parallelogram. Must be positive.

        a : int | float | None
            The length of one side of the parallelogram. Must be positive if provided.

        phi : int | float | None
            The angle in degrees opposite the base, adjacent to height. Must be between 0 and 180 if provided.

        Raises
        ------
        ValueError
            If base or height is not positive, or if angle is not in valid range.
        """
        super().__init__(num_of_sides=4)

        if base <= 0:
            raise ValueError("Base must be a positive number.")

        if height <= 0:
            raise ValueError("Height must be a positive number.")

        if a is not None and a <= 0:
            raise ValueError("Side 'a' must be a positive number.")

        if phi is not None and (phi <= 0 or phi >= 180):
            raise ValueError("Angle 'phi' must be between 0 and 180 degrees.")

        self.base = base
        self.height = height
        self.a = a
        self.phi = phi

    @override
    def perimeter(self) -> int | float:
        """
        Calculates and returns the perimeter of the parallelogram.

        Raises
        ------
        ValueError
            If neither side ``a`` nor angle ``phi`` is provided.
        """
        if self.a is not None:
            return 2 * (self.base + self.a)
        if self.phi is not None:
            side_b = self.height / math.sin(math.radians(self.phi))
            return 2 * (self.base + side_b)
            # return 2 * (self.base + self.height * math.cos(self.phi * math.pi / 180))
        raise ValueError("Side a or phi must be provided")

    @override
    def area(self) -> int | float:
        """Calculates and returns the area of the parallelogram."""
        return self.base * self.height


class Rhombus(Polygon):
    def __init__(self, d1: int | float, d2: int | float) -> None:
        """
        Initializes a Rhombus instance with specified diagonal lengths.

        Parameters
        ----------
        d1 : int | float
            The length of the first diagonal. Must be positive.

        d2 : int | float
            The length of the second diagonal. Must be positive.

        Raises
        ------
        ValueError
            If either diagonal length is not positive.
        """
        super().__init__(num_of_sides=4)

        if d1 <= 0:
            raise ValueError("Diagonal d1 must be a positive number.")

        if d2 <= 0:
            raise ValueError("Diagonal d2 must be a positive number.")

        self.d1 = d1
        self.d2 = d2

    @override
    def perimeter(self) -> float:
        """Calculates and returns the perimeter of the rhombus."""
        return 2 * math.sqrt(self.d1**2 + self.d2**2)

    @override
    def area(self) -> float:
        """Calculates and returns the area of the rhombus."""
        return (self.d1 * self.d2) / 2

    def side(self) -> float:
        """Calculates and returns the length of one side of the rhombus."""
        return self.perimeter() / self._num_of_sides  # type: ignore


class Trapezoid(Polygon):
    def __init__(
        self,
        a: int | float,
        b: int | float,
        c: int | float | None = None,
        d: int | float | None = None,
        h: int | float | None = None,
    ) -> None:
        """
        Initializes a Trapezoid instance with specified dimensions.

        Parameters
        ----------
        a : int | float
            The length of base 1. Must be positive.

        b : int | float
            The length of base 2. Must be positive.

        c : int | float | None
            The length of side 1. Must be positive if provided.

        d : int | float | None
            The length of side 2. Must be positive if provided.

        h : int | float | None
            The height of the trapezoid. Must be positive if provided.

        Raises
        ------
        ValueError
            If base lengths or height are not positive, or if both sides are not provided.
        """
        super().__init__(num_of_sides=4)

        if a <= 0:
            raise ValueError("Base 'a' must be a positive number.")

        if b <= 0:
            raise ValueError("Base 'b' must be a positive number.")

        if c is not None and c <= 0:
            raise ValueError("Side 'c' must be a positive number.")

        if d is not None and d <= 0:
            raise ValueError("Side 'd' must be a positive number.")

        if h is not None and h <= 0:
            raise ValueError("Height 'h' must be a positive number.")

        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.h = h

    @override
    def perimeter(self) -> int | float:
        """Calculates and returns the perimeter of the trapezoid."""
        if self.c is not None and self.d is not None:
            return sum([self.a, self.b, self.c, self.d])
        raise ValueError("'c' and 'd' must be provided to calculate perimeter.")

    @override
    def area(self) -> float:
        """Calculates and returns the area of the trapezoid."""
        if self.h is not None:
            return 0.5 * (self.a + self.b) * self.h
        raise ValueError("'h' must be provided to calculate area.")


# Class - 3D Shape
# ---------------------------------------------------------------------------
class Cube(ThreeDimensionShape):
    def __init__(self, side: int | float) -> None:
        """
        Initializes a Cube instance with a specified side length.

        Parameters
        ----------
        side : int | float
            The length of one side of the cube. Must be positive.

        Raises
        ------
        ValueError
            If the side length is not positive.
        """
        if side <= 0:
            raise ValueError("Side length must be a positive number.")

        self.side = side

    @override
    def surface_area(self) -> int | float:
        """Calculates and returns the surface area of the cube."""
        return 6 * self.side**2

    def surface_area_side(self) -> int | float:
        """Calculates and returns the surface area (side only) of the cube."""
        return 4 * self.side**2

    @override
    def volume(self) -> int | float:
        """Calculates and returns the volume of the cube."""
        return self.side**3

    def face_area(self) -> int | float:
        """Calculates and returns the area of one face of the cube."""
        return self.side**2

    def diagonal(self) -> float:
        """Calculates and returns the length of the space diagonal of the cube."""
        return self.side * math.sqrt(3)

    def scale(self, factor: int | float) -> None:
        """Scales the cube by a given factor."""
        if factor <= 0:
            raise ValueError("Scaling factor must be positive.")

        self.side *= factor

    @classmethod
    def from_surface_area(cls, surface_area: int | float) -> Self:
        """
        Creates a Cube instance from its surface area.

        Parameters
        ----------
        surface_area : int | float
            The surface area of the cube. Must be positive.

        Raises
        ------
        ValueError
            If surface area is not positive.
        """
        if surface_area <= 0:
            raise ValueError("Surface area must be positive.")

        # Calculate side length from surface area
        side = math.sqrt(surface_area / 6)
        return cls(side)

    @classmethod
    def from_volume(cls, volume: int | float) -> Self:
        """
        Creates a Cube instance from its volume.

        Parameters
        ----------
        volume : int | float
            The volume of the cube. Must be positive.

        Raises
        ------
        ValueError
            If volume is not positive.
        """
        if volume <= 0:
            raise ValueError("Volume must be positive.")

        # Calculate side length from volume
        side = volume ** (1 / 3)
        return cls(side)


class Cuboid(ThreeDimensionShape):
    def __init__(
        self,
        length: int | float,
        width: int | float,
        height: int | float,
    ) -> None:
        """
        Initializes a Cuboid instance with specified dimensions.

        Parameters
        ----------
        length : int | float
            The length of the cuboid. Must be positive.

        width : int | float
            The width of the cuboid. Must be positive.

        height : int | float
            The height of the cuboid. Must be positive.

        Raises
        ------
        ValueError
            If any dimension is not positive.
        """
        if length <= 0:
            raise ValueError("Length must be a positive number.")

        if width <= 0:
            raise ValueError("Width must be a positive number.")

        if height <= 0:
            raise ValueError("Height must be a positive number.")

        self.length = length
        self.width = width
        self.height = height

    @override
    def surface_area(self) -> int | float:
        """Calculates and returns the surface area of the cuboid."""
        res = 2 * (
            self.length * self.width
            + self.width * self.height
            + self.length * self.height
        )
        return res

    def surface_area_side(self) -> int | float:
        """Calculates and returns the surface area of the sides of the cuboid."""
        res = 2 * self.height * (self.length + self.width)
        return res

    @override
    def volume(self) -> int | float:
        """Calculates and returns the volume of the cuboid."""
        return self.length * self.width * self.height

    def diagonal(self) -> float:
        """Calculates and returns the length of the space diagonal of the cuboid."""
        return math.sqrt(self.length**2 + self.width**2 + self.height**2)

    def scale(self, factor: int | float) -> None:
        """Scales the dimensions of the cuboid by a given factor."""
        if factor <= 0:
            raise ValueError("Scaling factor must be positive.")

        self.length *= factor
        self.width *= factor
        self.height *= factor


class Sphere(ThreeDimensionShape):
    def __init__(self, radius: int | float) -> None:
        """
        Initializes a Sphere instance with the specified radius.

        Parameters
        ----------
        radius : int | float
            The radius of the sphere. Must be positive.

        Raises
        ------
        ValueError
            If radius is not positive.
        """
        if radius <= 0:
            raise ValueError("Radius must be a positive number.")

        self.radius = radius

    @override
    def surface_area(self) -> float:
        """Calculates and returns the surface area of the sphere."""
        return 4 * math.pi * self.radius**2

    @override
    def volume(self) -> float:
        """Calculates and returns the volume of the sphere."""
        return (4 / 3) * math.pi * self.radius**3

    @classmethod
    def from_volume(cls, volume: int | float) -> Self:
        """
        Creates a Sphere instance from its volume.

        Parameters
        ----------
        volume : int | float
            The volume of the sphere. Must be positive.

        Raises
        ------
        ValueError
            If volume is not positive.
        """
        if volume <= 0:
            raise ValueError("Volume must be a positive number.")

        radius = ((3 * volume) / (4 * math.pi)) ** (1 / 3)
        return cls(radius)

    @classmethod
    def from_surface_area(cls, surface_area: int | float) -> Self:
        """
        Creates a Sphere instance from its surface area.

        Parameters
        ----------
        surface_area : int | float
            The surface area of the sphere. Must be positive.

        Raises
        ------
        ValueError
            If surface area is not positive.
        """
        if surface_area <= 0:
            raise ValueError("Surface area must be a positive number.")

        radius = math.sqrt(surface_area / (4 * math.pi))
        return cls(radius)


class HemiSphere(ThreeDimensionShape):
    def __init__(self, radius: int | float) -> None:
        """
        Initializes a HemiSphere instance with the specified radius.

        Parameters
        ----------
        radius : int | float
            The radius of the hemisphere. Must be positive.

        Raises
        ------
        ValueError
            If the radius is not positive.
        """
        if radius <= 0:
            raise ValueError("Radius must be a positive number.")

        self.radius = radius

    @override
    def surface_area(self) -> float:
        """
        Calculates and returns the total surface area of the hemisphere.
        """
        return 3 * math.pi * self.radius**2

    def surface_area_curved(self) -> float:
        """
        Calculates and returns the curved surface area of the hemisphere.
        """
        return 2 * math.pi * self.radius**2

    def surface_area_base(self) -> float:
        """
        Calculates and returns the area of the base (circular) of the hemisphere.
        """
        return math.pi * self.radius**2

    @override
    def volume(self) -> float:
        """
        Calculates and returns the volume of the hemisphere.
        """
        return (2 / 3) * math.pi * self.radius**3

    @classmethod
    def from_volume(cls, volume: int | float) -> Self:
        """
        Creates a HemiSphere instance from its volume.

        Parameters
        ----------
        volume : int | float
            The volume of the hemisphere. Must be positive.

        Raises
        ------
        ValueError
            If the volume is not positive.
        """
        if volume <= 0:
            raise ValueError("Volume must be positive.")

        radius = (3 * volume / (2 * math.pi)) ** (1 / 3)
        return cls(radius)

    @classmethod
    def from_surface_area(cls, surface_area: int | float) -> Self:
        """
        Creates a HemiSphere instance from its total surface area.

        Parameters
        ----------
        surface_area : int | float
            The total surface area of the hemisphere. Must be positive.

        Raises
        ------
        ValueError
            If the surface area is not positive.
        """
        if surface_area <= 0:
            raise ValueError("Surface area must be positive.")

        radius = math.sqrt(surface_area / (3 * math.pi))
        return cls(radius)


class Cylinder(ThreeDimensionShape):
    def __init__(self, radius: int | float, height: int | float) -> None:
        """
        Initializes a Cylinder instance with the specified radius and height.

        Parameters
        ----------
        radius : int | float
            The radius of the cylinder. Must be positive.

        height : int | float
            The height of the cylinder. Must be positive.

        Raises
        ------
        ValueError
            If radius or height is not positive.
        """
        if radius <= 0:
            raise ValueError("Radius must be a positive number.")

        if height <= 0:
            raise ValueError("Height must be a positive number.")

        self.radius = radius
        self.height = height

    @override
    def surface_area(self) -> float:
        """Calculates and returns the total surface area of the cylinder."""
        res = 2 * math.pi * self.radius * (self.radius + self.height)
        return res

    def surface_area_curved(self) -> float:
        """Calculates and returns the curved surface area of the cylinder."""
        res = 2 * math.pi * self.radius * self.height
        return res

    @override
    def volume(self) -> float:
        """Calculates and returns the volume of the cylinder."""
        return math.pi * self.radius**2 * self.height



"""
Test: Version

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

from itertools import product

import pytest

from absfuyu.version import Bumper, ReleaseLevel, ReleaseOption, Version


# Version
class TestVersion:
    @pytest.mark.parametrize(
        ["major", "minor", "patch", "serial"], [(1, 0, 0, 5), ("1", "0", "0", "8")]
    )
    def test_init(
        self,
        major: int | str,
        minor: int | str,
        patch: int | str,
        serial: int | str,
    ) -> None:
        try:
            _ = Version(major, minor, patch, ReleaseLevel.DEV, serial)
            assert True
        except Exception as err:
            # assert False
            raise AssertionError(err)

    @pytest.mark.parametrize("data", [list(), dict(), set(), tuple()])
    def test_init_error(self, data) -> None:
        with pytest.raises(TypeError) as excinfo:
            _ = Version(data, data, data)
        assert str(excinfo.value)

    @pytest.mark.parametrize("tuple_", [(1, 0, 0), (1, 0, 0, "dev", 0)])
    def test_from_tuple(self, tuple_: tuple) -> None:
        try:
            _ = Version.from_tuple(tuple_)
            assert True
        except Exception as err:
            # assert False
            raise AssertionError(err)

    def test_from_tuple_error(self) -> None:
        with pytest.raises(ValueError) as excinfo:
            _ = Version.from_tuple((1, 0))
        assert str(excinfo.value)

    @pytest.mark.parametrize(
        "str_", ["1.0.0 ", " 100.8.48", "1.5.2.dev2", "1.2.4.rc8456", "1.9.4.final2"]
    )
    def test_from_str(self, str_: str) -> None:
        try:
            _ = Version.from_str(str_)
            assert True
        except Exception as err:
            # assert False
            raise AssertionError(err)

    @pytest.mark.parametrize("str_", ["1.0.0s", "5555.1.fnl5"])
    def test_from_str_error(self, str_: str) -> None:
        with pytest.raises(ValueError) as excinfo:
            _ = Version.from_str(str_)
        assert str(excinfo.value)


# Bumper
def list_of_bumper_bump_test() -> list[
    tuple[tuple[int, int, int, str, int], str, str, str]
]:
    bumper_list: list[tuple[int, int, int, str, int]] = [
        (1, 0, 0, ReleaseLevel.FINAL, 0),
        (1, 3, 2, ReleaseLevel.FINAL, 0),
        (2, 4, 6, ReleaseLevel.RC, 0),
        (2, 4, 6, ReleaseLevel.DEV, 0),
    ]
    release_options: list[str] = [
        ReleaseOption.MAJOR,
        ReleaseOption.MINOR,
        ReleaseOption.PATCH,
    ]
    release_levels: list[str] = [ReleaseLevel.FINAL, ReleaseLevel.DEV, ReleaseLevel.RC]
    merged = list(product(bumper_list, release_options, release_levels))
    # TODO: Improve this
    result_list: list[str] = [
        # 1.0.0
        "2.0.0",  # major final
        "2.0.0.dev0",  # major dev
        "2.0.0.rc0",  # major rc
        "1.1.0",  # minor final
        "1.1.0.dev0",  # minor dev
        "1.1.0.rc0",  # minor rc
        "1.0.1",  # patch final
        "1.0.1.dev0",  # patch dev
        "1.0.1.rc0",  # patch rc
        # 1.3.2
        "2.0.0",  # major
        "2.0.0.dev0",
        "2.0.0.rc0",
        "1.4.0",  # minor
        "1.4.0.dev0",
        "1.4.0.rc0",
        "1.3.3",  # patch
        "1.3.3.dev0",
        "1.3.3.rc0",
        # 2.4.6.rc0
        "2.4.6",  # major
        "3.0.0.dev0",
        "2.4.6.rc1",
        "2.4.6",  # minor
        "2.5.0.dev0",
        "2.4.6.rc1",
        "2.4.6",  # patch
        "2.4.7.dev0",
        "2.4.6.rc1",
        # 2.4.6.dev0
        "2.4.6",  # major
        "2.4.6.dev1",
        "2.4.6.rc0",
        "2.4.6",  # minor
        "2.4.6.dev1",
        "2.4.6.rc0",
        "2.4.6",  # patch
        "2.4.6.dev1",
        "2.4.6.rc0",
    ]
    if not len(result_list) == len(merged):
        # Check length
        raise ValueError("length of result_list is not in correct length")
    out: list[tuple[Bumper, str, str, str]] = [
        (x[0], x[1], x[2], result_list[i]) for i, x in enumerate(merged)
    ]
    return out


class TestBumper:
    @pytest.mark.parametrize(
        "bumper, release_option, release_channel, desired_output",
        list_of_bumper_bump_test(),
    )
    def test_bump(
        self,
        bumper: tuple[int, int, int, str, int],
        release_option: str,
        release_channel: str,
        desired_output: str,
    ) -> None:
        bumper: Bumper = Bumper.from_tuple(bumper)
        bumper.bump(option=release_option, channel=release_channel)
        assert str(bumper) == desired_output



"""
Absfuyu: Core
-------------
Dummy functions when other libraries are unvailable

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    "tqdm",
    "unidecode",
    "dummy_function",
]


# Library
# ---------------------------------------------------------------------------
from importlib import import_module

# Wrapper
# ---------------------------------------------------------------------------
# tqdm wrapper
try:
    _tqdm = import_module("tqdm")
    tqdm = getattr(_tqdm, "tqdm")  # noqa
except (ImportError, AttributeError):

    def tqdm(iterable, *args, **kwargs):
        """
        Dummy tqdm function,
        install package ``tqdm`` to fully use this feature
        """
        return iterable


# unidecode wrapper
try:
    _unidecode = import_module("unidecode")
    unidecode = getattr(_unidecode, "unidecode")  # noqa
except (ImportError, AttributeError):

    def unidecode(*args, **kwargs):
        """
        Dummy unidecode function,
        install package ``unidecode`` to fully use this feature
        """
        return args[0]


# dummy
def dummy_function(*args, **kwargs):
    """This is a dummy function"""
    if args:
        return args[0]
    if kwargs:
        return kwargs[list(kwargs)[0]]
    return None



"""
Test: Data extension - IntExt

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import pytest

from absfuyu.dxt import IntExt


@pytest.fixture
def num_a():
    return IntExt(5)


@pytest.fixture
def num_b():
    return IntExt(10)


@pytest.fixture
def num_prime():
    return IntExt(79)


@pytest.mark.abs_dxt
class TestIntExt:
    """absfuyu.dxt.IntExt"""

    # operation
    def test_operation(self, num_a: IntExt, num_b: IntExt) -> None:
        assert num_a + num_b == 15  # add
        assert num_a - num_b == -5  # subtract
        assert num_a * num_b == 50  # multiply
        assert num_a / num_b == 0.5  # divide
        assert (num_a > num_b) is False  # comparison

    # binary
    @pytest.mark.parametrize(
        ["number", "output"], [(10, "1010"), (10, format(10, "b"))]
    )
    def test_to_binary(self, number: int, output: str) -> None:
        instance: IntExt = IntExt(number)
        assert instance.to_binary() == output

    # reverse
    @pytest.mark.parametrize(["number", "output"], [(10, 1), (5, 5), (79, 97)])
    def test_to_reverse(self, number: int, output: int) -> None:
        instance: IntExt = IntExt(number)
        assert instance.reverse() == output

    # prime
    @pytest.mark.parametrize(["number", "output"], [(79, True), (33, False)])
    def test_is_prime(self, number: int, output: bool) -> None:
        instance: IntExt = IntExt(number)
        assert instance.is_prime() is output

    @pytest.mark.parametrize(["number", "output"], [(79, True), (53, False)])
    def test_is_twisted_prime(self, number: int, output: bool) -> None:
        instance: IntExt = IntExt(number)
        assert instance.is_twisted_prime() is output

    @pytest.mark.parametrize(["number", "output"], [(797, True), (79, False)])
    def test_is_palindromic_prime(self, number: int, output: bool) -> None:
        instance: IntExt = IntExt(number)
        assert instance.is_palindromic_prime() is output

    # perfect
    @pytest.mark.parametrize(["number", "output"], [(28, True), (22, False)])
    def test_is_perfect(self, number: int, output: bool) -> None:
        instance: IntExt = IntExt(number)
        assert instance.is_perfect() is output

    # narcissistic
    @pytest.mark.parametrize(["number", "output"], [(371, True), (46, False)])
    def test_is_narcissistic(self, number: int, output: bool) -> None:
        instance: IntExt = IntExt(number)
        assert instance.is_narcissistic() is output

    # palindromic
    @pytest.mark.parametrize(["number", "output"], [(12321, True), (1231, False)])
    def test_is_palindromic(self, number: int, output: bool) -> None:
        instance: IntExt = IntExt(number)
        assert instance.is_palindromic() is output

    # degree
    def test_convert_degree(self, num_a: IntExt) -> None:
        assert num_a.to_celcius_degree() == -15.0
        assert num_a.to_fahrenheit_degree() == 41.0

    # even
    @pytest.mark.parametrize(["number", "output"], [(2, True), (3, False)])
    def test_is_even(self, number: int, output: bool) -> None:
        instance: IntExt = IntExt(number)
        assert instance.is_even() is output

    # lcm
    def test_lcm(self, num_a: IntExt) -> None:
        assert num_a.lcm(6) == 30

    # gcd
    def test_gcd(self, num_a: IntExt) -> None:
        assert num_a.gcd(25) == 5

    # add_to_one_digit
    def test_add_to_one_digit(self, num_prime: IntExt) -> None:
        assert num_prime.add_to_one_digit() == 7

    @pytest.mark.parametrize(["number", "output"], [(1091, 11), (994, 22)])
    def test_add_to_one_digit_2(self, number: int, output: int) -> None:
        instance = IntExt(number)
        assert instance.add_to_one_digit(master_number=True) == output

    # analyze
    def test_analyze(self) -> None:
        assert IntExt(51564).analyze()

    # prime factor
    def test_prime_factor(self) -> None:
        assert IntExt(884652).prime_factor(short_form=False) == [2, 2, 3, 73721]

    # divisible_list
    def test_divisible_list(self) -> None:
        assert IntExt(884652).divisible_list() == [
            1,
            2,
            3,
            4,
            6,
            12,
            73721,
            147442,
            221163,
            294884,
            442326,
            884652,
        ]

    # split
    @pytest.mark.parametrize(
        ["parts", "output"],
        [
            (-10, [5]),
            (1, [5]),
            (2, [2, 3]),
            (5, [1, 1, 1, 1, 1]),
            (6, [0, 1, 1, 1, 1, 1]),
        ],
    )
    def test_split(self, num_a: IntExt, parts: int, output: list[int]) -> None:
        assert num_a.split(parts=parts) == output



"""
Absfuyu: Core
-------------
Pre-defined typing

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    # Type
    "T",
    "T_co",
    "T_contra",
    "KT",
    "VT",
    "P",
    "R",
    "_CALLABLE",
    "CT",
    "N",
    "_Number",
    "override",
    # Protocol
    "SupportsShowMethods",
]


# Library
# ---------------------------------------------------------------------------
from collections.abc import Callable
from typing import Any, ParamSpec, Protocol, TypeVar, overload

try:
    from typing import override  # type: ignore
except ImportError:
    from absfuyu.core.decorator import dummy_decorator as override


# Type
# ---------------------------------------------------------------------------
# Types where neither is possible are invariant
# Type safety must be strict
T = TypeVar("T")  # Type invariant
# Type variables that are covariant can be substituted
# with a more specific type without causing errors
# Safe to treat a Box[int] as a Box[object]
T_co = TypeVar("T_co", covariant=True)  # Type covariant
# Type variables that are contravariant can be substituted
# with a more general type without causing errors
# Safe to use a Box[object] where a Box[int] is needed
T_contra = TypeVar("T_contra", contravariant=True)  # Type contravariant

KT = TypeVar("KT")
VT = TypeVar("VT")

# Callable
P = ParamSpec("P")  # Parameter type
R = TypeVar("R")  # Return type - Can be anything
_CALLABLE = Callable[P, R]

# Class type - Can be any subtype of `type`
CT = TypeVar("CT", bound=type)

# Number type
N = TypeVar("N", int, float)  # Must be int or float
_Number = int | float


# Protocol
# ---------------------------------------------------------------------------
class SupportsShowMethods(Protocol):
    """
    Support class with ``show_all_methods()``
    and ``show_all_properties()`` method
    from ``absfuyu.core.basclass.ShowAllMethodsMixin``
    """

    @overload
    @classmethod
    def show_all_methods(cls) -> dict[str, list[str]]: ...

    @overload
    @classmethod
    def show_all_methods(
        cls,
        print_result: bool = False,
        include_classmethod: bool = True,
        classmethod_indicator: str = "<classmethod>",
        include_staticmethod: bool = True,
        staticmethod_indicator: str = "<staticmethod>",
        include_private_method: bool = False,
    ) -> dict[str, list[str]]: ...

    @classmethod
    def show_all_methods(cls, *args, **kwargs) -> Any: ...

    @overload
    @classmethod
    def show_all_properties(cls) -> dict[str, list[str]]: ...

    @overload
    @classmethod
    def show_all_properties(
        cls, print_result: bool = False
    ) -> dict[str, list[str]]: ...

    @classmethod
    def show_all_properties(cls, *args, **kwargs) -> Any: ...


# Note
# ---------------------------------------------------------------------------
# Iterable                               : __iter__
# Iterator(Iterable)                     : __next__, __iter__
# Reversible(Iterable)                   : __reversed__, __iter__
# Sized                                  : __len__
# Container                              : __contains__
# Collection(Sized, Iterable, Container) : __len__, __iter__, __contains__
# Set(Collection)                        : __contains__, __iter__, __len__
# MutableSet(Set)                        : __contains__, __iter__, __len__
# Mapping(Collection)                    : __getitem__, __iter__, and __len__
# MutableMapping(Mapping)                : __getitem__, __setitem__, __delitem__,__iter__, __len__
# Sequence(Reversible, Collection)       : __reversed__, __len__, __iter__, __contains__
# MutableSequence(Sequence)              : __getitem__, __setitem__, __delitem__, __reversed__, __len__, __iter__, __contains__

# Iterable                               : str, dict, list, tuple, set
# Iterator(Iterable)                     : Generator
# Reversible(Iterable)                   : str, dict, list, tuple
# Sized                                  : str, dict, list, tuple, set
# Container                              : str, dict, list, tuple, set
# Collection(Sized, Iterable, Container) : str, dict, list, tuple, set
# Set(Collection)                        : set
# MutableSet(Set)                        : set
# Mapping(Collection)                    : dict
# MutableMapping(Mapping)                : dict
# Sequence(Reversible, Collection)       : str, list, tuple
# MutableSequence(Sequence)              : list

# __iter__: for <...> in <...>
# __len__: len(<...>)
# __contains__: if <...> in <...>



"""
Absfuyu: Data Analysis
----------------------
Data Analyst DataFrame

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "DADF",
    "DataAnalystDataFrameColumnMethodMixin",
    "DataAnalystDataFrameRowMethodMixin",
    "DataAnalystDataFrameInfoMixin",
    "DataAnalystDataFrameNAMixin",
    "DataAnalystDataFrameOtherMixin",
    "DataAnalystDataFrameDateMixin",
    "DataAnalystDataFrameExportMixin",
    "DataAnalystDataFrameCityMixin",
]


# Library
# ---------------------------------------------------------------------------
import random
import string
from collections.abc import Callable, Sequence
from datetime import datetime, timedelta
from typing import Any, Literal, Self, cast

import numpy as np
import pandas as pd
from xlsxwriter import Workbook
from xlsxwriter.worksheet import Worksheet

try:
    from typing import override  # type: ignore
except ImportError:
    from absfuyu.core.decorator import dummy_decorator as override

try:
    from unidecode import unidecode  # type: ignore
except ImportError:
    from absfuyu.core.dummy_func import dummy_function as unidecode

from absfuyu.core.baseclass import GetClassMembersMixin
from absfuyu.core.docstring import deprecated, versionadded, versionchanged
from absfuyu.extra.da.dadf_base import CityData
from absfuyu.extra.da.dadf_base import DataAnalystDataFrameBase as DFBase
from absfuyu.extra.da.dadf_base import SplittedDF
from absfuyu.logger import logger
from absfuyu.typings import R as _R
from absfuyu.typings import T as _T
from absfuyu.util import set_min_max


# Column method
# ---------------------------------------------------------------------------
class DataAnalystDataFrameColumnMethodMixin(DFBase):
    """
    Data Analyst ``pd.DataFrame`` - Column method

    - Rearrange rightmost column
    - Drop columns
    - Drop rightmost column
    - Add blank column
    - Split str column
    - Get column name unidecoded
    - Get column unidecoded
    """

    def rearrange_rightmost_column(
        self, insert_to_col: str, num_of_cols: int = 1
    ) -> Self:
        """
        Move right-most columns to selected position

        Parameters
        ----------
        insert_to_col : str
            Name of the column that the right-most column will be moved next to

        num_of_cols : int
            Number of columns moved, by default ``1``

        Returns
        -------
        Self
            Modified DataFrame


        Example:
        --------
        >>> df = DADF.sample_df(2)
        >>> df
             number  number_big number_range  missing_value      text       date
        0 -1.583590         756          700            NaN  eqklyckc 2023-05-20
        1  0.203968         167          100            NaN  wzrsxinb 2011-02-27
        >>> df.rearrange_rightmost_column("number")
             number       date  number_big number_range  missing_value      text
        0 -1.583590 2023-05-20         756          700            NaN  eqklyckc
        1  0.203968 2011-02-27         167          100            NaN  wzrsxinb
        """
        cols: list[str] = self.columns.to_list()  # List of columns
        num_of_cols = int(set_min_max(num_of_cols, min_value=1, max_value=len(cols)))
        col_index: int = cols.index(insert_to_col)
        new_cols: list[str] = (
            cols[: col_index + 1]
            + cols[-num_of_cols:]
            + cols[col_index + 1 : len(cols) - num_of_cols]
        )
        self = self.__class__(self[new_cols])
        return self

    def drop_columns(self, columns: Sequence[str]) -> Self:
        """
        Drop columns in DataFrame

        Parameters
        ----------
        columns : Iterable[str]
            List of columns need to drop

        Returns
        -------
        Self
            Modified DataFrame


        Example:
        --------
        >>> df = DADF.sample_df(2)
        >>> df
             number  number_big number_range  missing_value      text       date
        0 -0.283019         666          600            NaN  ztoeeblx 2022-11-13
        1  1.194725         939          900            NaN  fxardqvh 2005-08-04
        >>> df.drop_columns(["date", "text"])
             number  number_big number_range  missing_value
        0 -0.283019         666          600            NaN
        1  1.194725         939          900            NaN
        """
        for column in columns:
            try:
                self.drop(columns=[column], inplace=True)
            except KeyError:
                logger.debug(f"{column} column does not exist")
                # pass
        return self

    def drop_rightmost(self, num_of_cols: int = 1) -> Self:
        """
        Drop ``num_of_cols`` right-most columns

        Parameters
        ----------
        num_of_cols : int
            Number of columns to drop

        Returns
        -------
        Self
            Modified DataFrame


        Example:
        --------
        >>> df = DADF.sample_df(2)
        >>> df
             number  number_big number_range  missing_value      text       date
        0  0.851953         572          500              5  ncpbnzef 2020-08-15
        1  0.381643         595          500             53  iojogbgj 2011-12-04
        >>> df.drop_rightmost(5)
             number
        0  0.851953
        1  0.381643
        """
        # Restrain
        # if num_of_cols < 1:
        #     num_of_cols = 1
        # if num_of_cols > self.shape[1]:
        #     num_of_cols = self.shape[1]
        num_of_cols = int(
            set_min_max(num_of_cols, min_value=1, max_value=self.shape[1])
        )

        # Logic
        for _ in range(num_of_cols):
            self.drop(self.columns[len(self.columns) - 1], axis=1, inplace=True)
        return self

    @deprecated("5.1.0", reason="Use pd.DataFrame.assign(...) method instead")
    def add_blank_column(self, column_name: str, fill: Any = np.nan, /) -> Self:
        """
        [DEPRECATED] Add a blank column.

        E.g: Use `pd.DataFrame.assign(new_col=lambda x: x['old_col'])` instead

        Parameters
        ----------
        column_name : str
            Name of the column to add

        fill : Any
            Fill the column with data

        Returns
        -------
        Self
            Modified DataFrame
        """
        self[column_name] = [fill] * self.shape[0]
        return self

    @versionadded("5.2.0")  # No test cases
    def split_str_column(
        self,
        col: str,
        pattern: str = " ",
        *,
        n: int | None = None,
        regex: bool = False,
    ) -> Self:
        """
        Split column with dtype[str] into other columns.

        Parameters
        ----------
        col : str
            Column name

        pattern : str, optional
            Split pattern, by default ``" "``

        n : int | None, optional
            Split by how many times, by default ``None``

        regex : bool, optional
            Regex mode, by default ``False``

        Returns
        -------
        Self
            DataFrame


        Example:
        --------
        >>> df = DADF(DADF.sample_df(5)[["text"]])
        >>> df.split_str_column("text", "s"))
               text    text_0 text_1
        0  uwfzbsgj     uwfzb     gj
        1  lxlskayx       lxl   kayx
        2  fzgpzjtp  fzgpzjtp   None
        3  lxnytktz  lxnytktz   None
        4  onryaxtt  onryaxtt   None
        """
        if n is None:
            pass
        splited_data: pd.DataFrame = self[col].str.split(pat=pattern, n=n, expand=True, regex=regex)  # type: ignore
        num_of_splitted_cols = splited_data.shape[1]
        new_col_names = [f"{col}_{x}" for x in range(num_of_splitted_cols)]
        self[new_col_names] = splited_data
        return self

    @versionadded("5.12.0")  # No test cases
    def get_column_name_unidecoded(self, col_name: str, /, *, mode: Literal["start", "end", "in"] = "start") -> str:
        """
        Get column name from lowercase unidecode'd version name

        Parameters
        ----------
        col_name : str
            Column name to find

        mode : Literal["start", "end", "in"], optional
            Which mode to find, by default "start"
            - "start": str.startswith()
            - "end": str.endswith()
            - "in": if x in y

        Returns
        -------
        str
            Column name

        Raises
        ------
        ValueError
            Column not found
        """
        for x in self.columns.to_list():
            col_name_mod = cast(str, unidecode(x.strip().lower()))
            if mode == "start":
                if col_name_mod.startswith(col_name):
                    return x
            elif mode == "end":
                if col_name_mod.endswith(col_name):
                    return x
            elif mode == "in":
                if col_name_mod in col_name:
                    return x

        raise ValueError(f"Column not found: {col_name}")

    @versionadded("5.12.0")  # No test cases
    def get_column_unidecoded(self, col_name: str, /, *, mode: Literal["start", "end", "in"] = "start") -> pd.Series:
        """
        Get column from lowercase unidecode'd version column name

        Parameters
        ----------
        col_name : str
            Column name to find

        mode : Literal["start", "end", "in"], optional
            Which mode to find, by default "start"
            - "start": str.startswith()
            - "end": str.endswith()
            - "in": if x in y

        Returns
        -------
        Series
            Column data
        """
        return self[self.get_column_name_unidecoded(col_name, mode=mode)]


# Row method
# ---------------------------------------------------------------------------
class DataAnalystDataFrameRowMethodMixin(DFBase):
    """
    Data Analyst ``pd.DataFrame`` - Row method

    - Get different rows
    - Add blank row
    """

    @versionadded("4.0.0")
    def get_different_rows(self, other: Self | pd.DataFrame) -> Self:
        """
        Subtract DataFrame to find the different rows

        Parameters
        ----------
        other : Self | pd.DataFrame
            DataFrame to subtract

        Returns
        -------
        Self
            Different row DataFrame


        Example:
        --------
        >>> df1 = DADF({"A": [1, 2, 3, 4], "B": [5, 6, 7, 8]})
        >>> df2 = DADF({"A": [1, 2, 3, 4], "B": [7, 6, 6, 8]})
        >>> df1.get_different_rows(df2)
           A  B
        0  1  7
        2  3  6
        """
        df = self.copy()
        out = (
            df.merge(other, indicator=True, how="right")
            .query("_merge=='right_only'")
            .drop("_merge", axis=1)
        )
        return self.__class__(out)

    @versionadded("5.7.0")
    def add_blank_row(self, fill: Any = np.nan, /) -> Self:
        """
        Add a new row to the end of a DataFrame.

        Parameters
        ----------
        fill : Any, default np.nan
            Value to fill in the new row (e.g., np.nan, None, "", 0).

        Returns
        -------
        Self
            DataFrame with the new row appended.
        """
        # Create a dict with all columns filled with fill
        new_row = {col: fill for col in self.columns}
        self.loc[len(self)] = new_row  # type: ignore
        return self


# Info
# ---------------------------------------------------------------------------
class DataAnalystDataFrameInfoMixin(DFBase):
    """
    Data Analyst ``pd.DataFrame`` - Info

    - Quick info
    - Quick describe
    - Show distribution
    - Threshold filter
    """

    # Quick info
    @versionadded("3.2.0")
    def qinfo(self) -> str:
        """
        Show quick infomation about DataFrame

        Example:
        --------
        >>> DADF.sample_df().qinfo()
        Dataset Information:
        - Number of Rows: 100
        - Number of Columns: 6
        - Total observation: 600
        - Missing value: 13 (2.17%)

        Column names:
        ['number', 'number_big', 'number_range', 'missing_value', 'text', 'date']
        """
        missing_values = self.isnull().sum().sum()
        total_observation = self.shape[0] * self.shape[1]
        mv_rate = missing_values / total_observation * 100
        info = (
            f"Dataset Information:\n"
            f"- Number of Rows: {self.shape[0]:,}\n"
            f"- Number of Columns: {self.shape[1]:,}\n"
            f"- Total observation: {total_observation:,}\n"
            f"- Missing value: {missing_values:,} ({mv_rate:.2f}%)\n\n"
            f"Column names:\n{self.columns.to_list()}"
        )
        return info

    @override
    def describe(self, percentiles=None, include=None, exclude=None) -> Self:  # type: ignore
        """pd.DataFrame.describe() override"""
        return self.__class__(super().describe(percentiles, include, exclude))  # type: ignore [no-any-return]

    # Quick describe
    @versionadded("3.2.0")
    def qdescribe(self) -> Self:
        """
        Quick ``describe()`` that exclude ``object`` and ``datetime`` dtype

        Returns
        -------
        Self
            Modified DataFrame


        Example:
        --------
        >>> DADF.sample_df().qdescribe()
                   number  number_big  missing_value
        count  100.000000  100.000000      48.000000
        mean    -0.052935  586.750000      22.916667
        std      0.954170  237.248596      11.987286
        min     -2.392952  105.000000       3.000000
        25%     -0.738311  407.500000      13.000000
        50%     -0.068014  607.000000      23.500000
        75%      0.614025  790.250000      36.000000
        max      2.512533  988.000000      42.000000
        """
        return self.__class__(  # type: ignore [no-any-return]
            self[self.select_dtypes(exclude=["object", "datetime"]).columns].describe()
        )

    @versionadded("3.2.0")
    def show_distribution(
        self,
        column_name: str,
        dropna: bool = True,
        *,
        show_percentage: bool = True,
        percentage_round_up: int = 2,
    ) -> Self:
        """
        Show distribution of a column

        Parameters
        ----------
        column_name : str
            Column to show distribution

        dropna : bool
            Count N/A when ``False``
            (Default: ``True``)

        show_percentage : bool
            Show proportion in range 0% - 100% instead of [0, 1]
            (Default: ``True``)

        percentage_round_up : int
            Round up to which decimals
            (Default: ``2``)

        Returns
        -------
        Self
            Distribution DataFrame


        Example:
        --------
        >>> DADF.sample_df().show_distribution("number_range")
          number_range  count  percentage
        0          900     16        16.0
        1          700     15        15.0
        2          300     12        12.0
        3          200     12        12.0
        4          400     11        11.0
        5          600     11        11.0
        6          800     10        10.0
        7          100      9         9.0
        8          500      4         4.0
        """
        out = self[column_name].value_counts(dropna=dropna).to_frame().reset_index()
        if show_percentage:
            out["percentage"] = (out["count"] / self.shape[0] * 100).round(
                percentage_round_up
            )
        else:
            out["percentage"] = (out["count"] / self.shape[0]).round(
                percentage_round_up
            )
        return self.__class__(out)

    @deprecated("5.1.0", reason="Rework THIS")
    def threshold_filter(
        self,
        destination_column: str,
        threshold: int | float = 10,
        *,
        top: int | None = None,
        replace_with: Any = "Other",
    ) -> Self:
        """
        Filter out percentage of data that smaller than the ``threshold``,
        replace all of the smaller data to ``replace_with``.
        As a result, pie chart is less messy.

        Parameters
        ----------
        destination_column : str
            Column to be filtered

        threshold : int | float
            Which percentage to cut-off
            (Default: 10%)

        top : int
            Only show top ``x`` categories in pie chart
            (replace threshold mode)
            (Default: ``None``)

        replace_with : Any
            Replace all of the smaller data with specified value

        Returns
        -------
        Self
            Modified DataFrame
        """
        # Clean
        try:
            self[destination_column] = self[
                destination_column
            ].str.strip()  # Remove trailing space
        except Exception:
            pass

        # Logic
        col_df = self.show_distribution(destination_column)

        # Rename
        if top is not None:
            list_of_keep: list = (
                col_df[destination_column]
                .head(set_min_max(top - 1, min_value=1, max_value=col_df.shape[0]))  # type: ignore
                .to_list()
            )
            # logger.debug(list_of_keep)
        else:
            list_of_keep = col_df[col_df["percentage"] >= threshold][
                destination_column
            ].to_list()  # values that will not be renamed
        self[f"{destination_column}_filtered"] = self[destination_column].apply(
            lambda x: replace_with if x not in list_of_keep else x
        )

        # Return
        return self


# Missing value
# ---------------------------------------------------------------------------
class DataAnalystDataFrameNAMixin(DFBase):
    """
    Data Analyst ``pd.DataFrame`` - Missing value

    - Fill missing values
    - Get missing values
    - Split N/A
    - Apply not null
    - Apply not null row
    """

    def fill_missing_values(
        self, column_name: str, fill: Any = np.nan, *, fill_when_not_exist: Any = np.nan
    ) -> Self:
        """
        Fill missing values in specified column

        Parameters
        ----------
        column_name : str
            Column name

        fill : Any
            Fill the missing values with, by default ``np.nan``

        fill_when_not_exist : Any
            When ``column_name`` does not exist,
            create a new column and fill with
            ``fill_when_not_exist``, by default ``np.nan``

        Returns
        -------
        Self
            Modified DataFrame


        Example:
        --------
        >>> df = DADF.sample_df(2)
        >>> df
             number  number_big number_range  missing_value      text       date
        0  0.174303         926          900            NaN  tenkiakh 2006-09-08
        1  0.305137         140          100            NaN  jzuddamf 2012-04-04
        >>> df.fill_missing_values("missing_value", 0)
             number  number_big number_range  missing_value      text       date
        0  0.174303         926          900            0.0  tenkiakh 2006-09-08
        1  0.305137         140          100            0.0  jzuddamf 2012-04-04
        >>> df.fill_missing_values("missing_column", 0, fill_when_not_exist=0)
             number  number_big number_range  missing_value      text       date  missing_column
        0  0.174303         926          900            0.0  tenkiakh 2006-09-08               0
        1  0.305137         140          100            0.0  jzuddamf 2012-04-04               0
        """
        try:
            self[column_name] = self[column_name].fillna(fill)
        except KeyError:
            if getattr(self, "add_blank_column", None) is not None:
                # Compatible with DataAnalystDataFrameColumnMethodMixin
                self.add_blank_column(column_name, fill_when_not_exist)  # type: ignore
        return self

    def get_missing_values(
        self, hightlight: bool = True, *, percentage_round_up: int = 2
    ) -> Self:
        """
        Get a DataFrame contains count of missing values for each column

        Parameters
        ----------
        hightlight : bool
            Shows only columns with missing values when ``True``, by default ``True``

        percentage_round_up : int
            Round up to which decimals, by default ``2``

        Returns
        -------
        Self
            Missing value DataFrame


        Example:
        --------
        >>> DADF.sample_df(152).get_missing_values()
                       Num of N/A  Percentage
        missing_value          42       27.63
        """
        # Check for missing value
        df_na = self.isnull().sum().sort_values(ascending=False)
        if hightlight:
            out = df_na[df_na != 0].to_frame()
        else:
            out = df_na.to_frame()
        out.rename(columns={0: "Num of N/A"}, inplace=True)
        out["Percentage"] = (out["Num of N/A"] / self.shape[0] * 100).round(
            percentage_round_up
        )

        # logger.debug(
        #     f"Percentage of N/A over entire DF: "
        #     f"{(self.isnull().sum().sum() / (self.shape[0] * self.shape[1]) * 100).round(percentage_round_up)}%"
        # )
        return self.__class__(out)

    @versionadded("3.1.0")
    def split_na(self, by_column: str) -> SplittedDF:
        """
        Split DataFrame into 2 parts:
            - Without missing value in specified column
            - With missing value in specified column

        Parameters
        ----------
        by_column : str
            Split by column

        Returns
        -------
        SplittedDF
            Splitted DataFrame


        Example:
        --------
        >>> DADF.sample_df(10).split_na("missing_value")
        SplittedDF(
            df=     number  number_big number_range  missing_value      text       date
        0         0.643254         690          600            3.0  cinvofwj 2018-08-15
        2         0.499345         255          200           13.0  jasifzez 2005-06-01
        3        -1.727036         804          800           38.0  esxjmger 2009-07-24
        4         0.873058         690          600           32.0  htewfpld 2022-07-22
        5        -2.389884         442          400           30.0  hbcnfogu 2006-02-25
        8         0.264584         432          400            2.0  ejbvbmwn 2013-05-11
        9         0.813655         137          100           20.0  oecttada 2024-11-22,
            df_na=     number  number_big number_range  missing_value      text       date
        1           -0.411354         363          300            NaN  juzecani 2014-12-02
        6           -0.833857         531          500            NaN  ybnntryh 2023-11-03
        7            1.355589         472          400            NaN  zjltghjr 2024-10-09
        )
        """
        out = SplittedDF(
            # df=self[~self[by_column].isna()],  # DF
            df=self[self[by_column].notna()],  # DF
            df_na=self[self[by_column].isna()],  # DF w/o NA
        )
        return out

    @versionadded("5.1.0")
    def apply_notnull(self, col: str, callable: Callable[[Any], _R]) -> Self:
        """
        Only apply callable to not NaN value in column

        Parameters
        ----------
        col : str
            Column to apply

        callable : Callable[[Any], _R]
            Callable

        Returns
        -------
        Self
            Applied DataFrame


        Example:
        --------
        >>> DADF.sample_df(5).apply_notnull("missing_value", lambda _: "REPLACED")
             number  number_big number_range missing_value      text       date
        0  0.852218         157          100      REPLACED  dqzxaxxs 2006-03-08
        1  1.522428         616          600           NaN  mivkaooe 2018-12-27
        2  0.108506         745          700      REPLACED  qanwwjet 2005-07-14
        3 -1.435079         400          400      REPLACED  ywahcasi 2024-05-20
        4  0.118993         861          800      REPLACED  saoupuby 2019-04-28
        """
        self[col] = self[col].apply(lambda x: callable(x) if pd.notnull(x) else x)  # type: ignore
        return self

    @versionadded("5.1.0")  # type: ignore
    def apply_notnull_row(
        self,
        apply_when_null: Callable[[Any], _R] | _T | None = None,
        apply_when_not_null: Callable[[Any], _R] | _T | None = None,
        col_name: str | None = None,
    ) -> Self:
        """
        Apply to DataFrame's row with missing value.

        Parameters
        ----------
        apply_when_null : Callable[[Any], R] | T | None, optional
            Callable or Any, by default ``None``: returns if entire row is not null

        apply_when_not_null : Callable[[Any], R] | T | None, optional
            Callable or Any, by default ``None``: returns if entire row is not null

        col_name : str | None, optional
            Output column name, by default ``None`` (uses custom name)

        Returns
        -------
        Self
            Modified DataDrame


        Example:
        --------
        >>> df = DADF({"A": [None, 2, 3, 4], "B": [1, None, 3, 4], "C": [None, 2, None, 4]})
        >>> df.apply_notnull_row()
             A    B    C  applied_row_null
        0  NaN  1.0  NaN             False
        1  2.0  NaN  2.0             False
        2  3.0  3.0  NaN             False
        3  4.0  4.0  4.0              True
        >>> df.apply_notnull_row(0, 1)
             A    B    C  applied_row_null
        0  NaN  1.0  NaN                 0
        1  2.0  NaN  2.0                 0
        2  3.0  3.0  NaN                 0
        3  4.0  4.0  4.0                 1
        >>> df.apply_notnull_row(lambda _: "n", lambda _: "y", col_name="mod")
             A    B    C mod
        0  NaN  1.0  NaN   n
        1  2.0  NaN  2.0   n
        2  3.0  3.0  NaN   n
        3  4.0  4.0  4.0   y
        """

        def apply_func(row: pd.Series):
            # Both None
            if apply_when_null is None and apply_when_not_null is None:
                return row.notnull().all()

            # When all values in row are not null
            if row.notnull().all():
                if callable(apply_when_not_null):
                    return apply_when_not_null(row)
                return apply_when_not_null

            # When any value in row is null
            if callable(apply_when_null):
                return apply_when_null(row)
            return apply_when_null

        # Column name
        cname = "applied_row_null" if col_name is None else col_name
        self[cname] = self.apply(apply_func, axis=1)  # type: ignore

        return self


# Other
# ---------------------------------------------------------------------------
class DataAnalystDataFrameOtherMixin(DFBase):
    """
    Data Analyst ``pd.DataFrame`` - Other method/Stuff

    - Merge left
    """

    @versionadded("4.0.0")
    def merge_left(
        self,
        other: Self | pd.DataFrame,
        on: str,
        columns: list[str] | None = None,
    ) -> Self:
        """
        Merge left of 2 DataFrame

        Parameters
        ----------
        other : Self | pd.DataFrame
            DataFrame to merge

        on : str
            Merge on which column

        columns : list[str] | None, optional
            Columns to take from other DataFrame, by default ``None``
            (Take all columns)

        Returns
        -------
        Self
            Merged DataFrame


        Example:
        --------
        >>> df1 = DADF({
        ...     "id": [1, 2, 5],
        ...     "name": ["Alice", "Bob", "Rich"],
        ...     "age": [20, 20, 20],
        ... })
        >>> df2 = DADF({
        ...     "id": [1, 2, 3],
        ...     "age": [25, 30, 45],
        ...     "department": ["HR", "IT", "PM"],
        ...     "salary": [50000, 60000, 55000],
        ... })
        >>> df1.merge_left(df2, on="id")
           id   name  age_x  age_y department   salary
        0   1  Alice     20   25.0         HR  50000.0
        1   2    Bob     20   30.0         IT  60000.0
        2   5   Rich     20    NaN        NaN      NaN
        >>> df1.merge_left(df2, on="id", columns=["salary"])
           id   name   age department   salary
        0   1  Alice  25.0         HR  50000.0
        1   2    Bob  30.0         IT  60000.0
        2   5   Rich   NaN        NaN      NaN
        """

        if columns is not None:
            current_col = [on]
            current_col.extend(columns)
            col = other.columns.to_list()
            cols = list(set(col) - set(current_col))

            if getattr(self, "drop_columns", None) is not None:
                # Compatible with DataAnalystDataFrameColumnMethodMixin
                self.drop_columns(cols)  # type: ignore

        out = self.merge(other, how="left", on=on)
        return self.__class__(out)


# Date
# ---------------------------------------------------------------------------
class DataAnalystDataFrameDateMixin(DFBase):
    """
    Data Analyst ``pd.DataFrame`` - Date

    - Add date column from month column
    - Add detail date
    - Delta date (How many days inbetween)
    """

    def add_date_from_month(self, month_column: str, *, col_name: str = "date") -> Self:
        """
        Add dummy ``date`` column from ``month`` column

        Parameters
        ----------
        month_column : str
            Month column

        col_name : str
            New date column name, by default: ``"date"``

        Returns
        -------
        Self
            Modified DataFrame


        Example:
        --------
        >>> df = (
        ...     DADF.sample_df(2)
        ...     .add_detail_date("date", mode="m")
        ...     .drop_columns(["date", "number", "number_range"])
        ... )
        >>> df
           number_big  missing_value      text  month
        0         755            NaN  lincgqzl      4
        1         907            NaN  gxltrjku     10
        >>> df.add_date_from_month("month")
           number_big  missing_value      text  month       date
        0         755            NaN  lincgqzl      4 2025-04-01
        1         907            NaN  gxltrjku     10 2025-10-01
        """
        _this_year = datetime.now().year
        self[col_name] = pd.to_datetime(
            f"{_this_year}-" + self[month_column].astype(int).astype(str) + "-1",
            format="%Y-%m-%d",
        )

        # Rearrange
        if getattr(self, "rearrange_rightmost_column", None) is not None:
            # Compatible with DataAnalystDataFrameColumnMethodMixin
            return self.rearrange_rightmost_column(month_column)  # type: ignore [no-any-return]
        return self

    def add_detail_date(self, date_column: str, mode: str = "dwmy") -> Self:
        """
        Add these columns from ``date_column``:
            - ``date`` (won't add if ``date_column`` value is ``"date"``)
            - ``day`` (overwrite if already exist)
            - ``week`` (overwrite if already exist)
            - ``month`` (overwrite if already exist)
            - ``year``  (overwrite if already exist)

        Parameters
        ----------
        date_column : str
            Date column

        mode : str
            | Detailed column to add
            | ``d``: day
            | ``w``: week number
            | ``m``: month
            | ``y``: year
            | (Default: ``"dwmy"``)

        Returns
        -------
        Self
            Modified DataFrame


        Example:
        --------
        >>> df = DADF.sample_df(2)
        >>> df
             number  number_big number_range  missing_value      text       date
        0  0.331195         902          900             20  fgyanxik 2021-10-18
        1 -0.877727         378          300             13  dqvaggjo 2007-03-06
        >>> df.add_detail_date("date")
             number  number_big number_range  missing_value      text       date  day  week  month  year
        0  0.331195         902          900             20  fgyanxik 2021-10-18   18    42     10  2021
        1 -0.877727         378          300             13  dqvaggjo 2007-03-06    6    10      3  2007
        """
        # Convert to datetime
        self["date"] = pd.to_datetime(self[date_column])

        # Logic
        col_counter = 0
        # self["weekday"] = self["day"].dt.isocalendar().day # Weekday
        if mode.find("d") != -1:
            logger.debug("Mode: 'day'")
            self["day"] = self["date"].dt.day
            col_counter += 1
        if mode.find("w") != -1:
            logger.debug("Mode: 'weekday'")
            self["week"] = self["date"].dt.isocalendar().week
            col_counter += 1
        if mode.find("m") != -1:
            logger.debug("Mode: 'month'")
            self["month"] = self["date"].dt.month
            col_counter += 1
        if mode.find("y") != -1:
            logger.debug("Mode: 'year'")
            self["year"] = self["date"].dt.year
            col_counter += 1

        # Return
        if getattr(self, "rearrange_rightmost_column", None) is not None:
            # Compatible with DataAnalystDataFrameColumnMethodMixin
            return self.rearrange_rightmost_column(date_column, col_counter)  # type: ignore [no-any-return]
        return self

    def delta_date(
        self,
        date_column: str,
        mode: Literal["now", "between_row"] = "now",
        *,
        col_name: str = "delta_date",
    ) -> Self:
        """
        Calculate date interval

        Parameters
        ----------
        date_column : str
            Date column

        mode : str
            | Mode to calculate
            | ``"between_row"``: Calculate date interval between each row
            | ``"now"``: Calculate date interval to current date
            | (Default: ``"now"``)

        col_name : str
            | New delta date column name
            | (Default: ``"delta_date"``)

        Returns
        -------
        Self
            Modified DataFrame


        Example:
        --------
        >>> df = DADF.sample_df(2)
        >>> df
             number  number_big number_range  missing_value      text       date
        0 -0.729988         435          400             21  xkrqqouf 2014-08-01
        1 -0.846031         210          200              5  rbkmiqxt 2024-07-10
        >>> df.delta_date("date")
             number  number_big number_range  missing_value      text       date  delta_date
        0 -0.729988         435          400             21  xkrqqouf 2014-08-01        3873
        1 -0.846031         210          200              5  rbkmiqxt 2024-07-10         242
        """
        if mode.lower().startswith("between_row"):
            dated = self[date_column].to_list()
            cal: list[timedelta] = []
            for i in range(len(dated)):
                if i == 0:
                    cal.append(dated[i] - dated[i])
                    # cal.append(relativedelta(dated[i], dated[i]))
                else:
                    cal.append(dated[i] - dated[i - 1])
                    # cal.append(relativedelta(dated[i], dated[i - 1]))
            self[col_name] = [x.days for x in cal]
        else:  # mode="now"
            self[col_name] = self[date_column].apply(
                lambda x: (datetime.now() - x).days
            )
        return self


# Export
# ---------------------------------------------------------------------------
class DataAnalystDataFrameExportMixin(DFBase):
    """
    Data Analyst ``pd.DataFrame`` - Export method

    - da_export
    """

    @versionchanged("5.8.0", "New parameter")
    def da_export(
        self,
        path: str,
        sheet_name: str = "Sheet1",
        *,
        auto_width: bool = True,
        cols_contain_centered_text: list[str] | None = None,
        cols_contain_number: list[str] | None = None,
        cols_contain_percentage: list[str] | None = None,
    ) -> None:
        """
        Export DataFrame with `xlsxwriter` engine

        Parameters
        ----------
        path : Path | str
            Path to export

        sheet_name : str, optional
            Sheet name, by default "Sheet1"

        auto_width : bool, optional
            Auto resize column width, by default ``True``

        cols_contain_centered_text : list[str] | None, optional
            Columns that contain centered text (Align center), by default None

        cols_contain_number : list[str] | None, optional
            Columns that contain number value (to format as number - int), by default None

        cols_contain_percentage : list[str] | None, optional
            Columns that contain percentage value (to format as percentage), by default None
        """

        # Using xlsxwriter engine
        with pd.ExcelWriter(path, engine="xlsxwriter") as writer:
            self.to_excel(writer, sheet_name=sheet_name, index=False, float_format="%.2f", na_rep="")

            # Format style
            workbook: Workbook = writer.book  # type: ignore
            header_fmt = workbook.add_format(
                {
                    "bold": True,
                    "text_wrap": True,
                    "border": 1,
                    "align": "center",
                    "valign": "vcenter",
                    # "bg_color": "#A0BEFD",
                }
            )
            number_fmt = workbook.add_format(
                {"num_format": "#,##0", "align": "center", "valign": "vcenter"}
            )  # 1,000,000
            percent_fmt = workbook.add_format({"num_format": "0.00%", "align": "center", "valign": "vcenter"})  # 1.00%
            text_fmt = workbook.add_format({"valign": "vcenter"})
            text_center_fmt = workbook.add_format({"align": "center", "valign": "vcenter"})

            # Format sheet
            worksheet: Worksheet = writer.sheets[sheet_name]

            # Format header - First row
            for col_num, value in enumerate(self.columns.values):
                worksheet.write(0, col_num, value, header_fmt)

            rules = [
                (cols_contain_number, number_fmt),
                (cols_contain_percentage, percent_fmt),
                (cols_contain_centered_text, text_center_fmt),
            ]

            # Auto width + col format
            for i, col in enumerate(self.columns):
                # Max str len of each column
                max_len = None if auto_width is None else max(self[col].astype(str).map(len).max(), len(col)) + 2
                worksheet.set_column(i, i, max_len)  # Set width

                # Format style
                fmt = text_fmt  # default
                for cols, f in rules:
                    if cols is not None and col in cols:
                        fmt = f
                        break
                worksheet.set_column(i, i, max_len, fmt)

            # if cols_contain_number is not None:
            #     for x in cols_contain_number:
            #         self[x] = pd.to_numeric(self[x], errors="coerce")


# City
# ---------------------------------------------------------------------------
class DataAnalystDataFrameCityMixin(DFBase):
    """
    Data Analyst ``pd.DataFrame`` - City

    - Convert city
    """

    def convert_city(
        self,
        city_column: str,
        city_list: list[CityData],
        *,
        mode: str = "ra",
    ) -> Self:
        """
        Get ``region`` and ``area`` of a city

        Parameters
        ----------
        city_column : str
            Column contains city data

        city_list : list[CityData]
            List of city in correct format
            (Default: ``None``)

        mode : str
            | Detailed column to add
            | ``r``: region
            | ``a``: area
            | (Default: ``"ra"``)

        Returns
        -------
        DataAnalystDataFrame
            Modified DataFrame
        """

        # Support function
        def _convert_city_support(value: str) -> CityData:
            for x in city_list:
                if x.city.lower().startswith(value.lower()):
                    return x
            return CityData(city=value, region=np.nan, area=np.nan)  # type: ignore

        # Convert
        col_counter = 0
        if mode.find("r") != -1:
            logger.debug("Mode: 'region'")
            self["region"] = self[city_column].apply(
                lambda x: _convert_city_support(x).region
            )
            col_counter += 1
        if mode.find("a") != -1:
            logger.debug("Mode: 'area'")
            self["area"] = self[city_column].apply(
                lambda x: _convert_city_support(x).area
            )
            col_counter += 1

        # Rearrange
        if getattr(self, "rearrange_rightmost_column", None) is not None:
            return self.rearrange_rightmost_column(city_column, col_counter)  # type: ignore [no-any-return]
        return self


# Main
# ---------------------------------------------------------------------------
class DADF(
    GetClassMembersMixin,
    DataAnalystDataFrameCityMixin,
    DataAnalystDataFrameExportMixin,
    DataAnalystDataFrameDateMixin,
    DataAnalystDataFrameOtherMixin,
    DataAnalystDataFrameNAMixin,
    DataAnalystDataFrameInfoMixin,
    DataAnalystDataFrameRowMethodMixin,
    DataAnalystDataFrameColumnMethodMixin,
):
    """
    Data Analyst ``pd.DataFrame``

    For a list of extra methods:
    >>> print(DADF.DADF_METHODS)
    """

    @classmethod
    @deprecated("5.1.0")
    @versionadded("3.2.0")
    def dadf_help(cls) -> list[str]:
        """
        Show all available method of DataAnalystDataFrame
        """
        list_of_method = list(set(dir(cls)) - set(dir(pd.DataFrame)))
        return sorted(list_of_method)

    @classmethod
    def sample_df(cls, size: int = 100) -> Self:
        """
        Create sample DataFrame

        Parameters
        ----------
        size : int
            Number of observations, by default ``100``

        Returns
        -------
        Self
            DataFrame with these columns:
            [number, number_big, number_range, missing_value, text, date]


        Example:
        --------
        >>> DataAnalystDataFrame.sample_df()
              number  number_big number_range  missing_value      text       date
        0  -2.089770         785          700            NaN  vwnlqoql 2013-11-20
        1  -0.526689         182          100           24.0  prjjcvqc 2007-04-13
        2  -1.596514         909          900            8.0  cbcpzlac 2023-05-24
        3   2.982191         989          900           21.0  ivwqwuvd 2022-04-28
        4   1.687803         878          800            NaN  aajtncum 2005-10-05
        ..       ...         ...          ...            ...       ...        ...
        95 -1.295145         968          900           16.0  mgqunkhi 2016-04-12
        96  1.296795         255          200            NaN  lwvytego 2014-05-10
        97  1.440746         297          200            5.0  lqsoykun 2010-04-03
        98  0.327702         845          800            NaN  leadkvsy 2005-08-05
        99  0.556720         981          900           36.0  bozmxixy 2004-02-22
        [100 rows x 6 columns]
        """
        # Restrain
        size = max(size, 1)

        # Number col
        df = cls(np.random.randn(size, 1), columns=["number"])
        df["number_big"] = [
            random.choice(range(100, 999)) for _ in range(size)
        ]  # Big number in range 100-999
        df["number_range"] = df["number_big"].apply(lambda x: str(x)[0] + "00")

        # Missing value col
        na_rate = random.randint(1, 99)
        d = [random.randint(1, 99) for _ in range(size)]
        df["missing_value"] = list(map(lambda x: x if x < na_rate else np.nan, d))
        # df["missing_value"] = [random.choice([random.randint(1, 99), np.nan]) for _ in range(observations)]

        # Text col
        df["text"] = [
            "".join([random.choice(string.ascii_lowercase) for _ in range(8)])
            for _ in range(size)
        ]

        # Random date col
        df["date"] = [
            datetime(
                year=random.randint(datetime.now().year - 20, datetime.now().year),
                month=random.randint(1, 12),
                day=random.randint(1, 28),
            )
            for _ in range(size)
        ]

        # Return
        return df


class DADF_WIP(DADF):
    """
    W.I.P - No test cases written
    """

    pass

if __name__ == "__main__":
    from pathlib import Path

    t = DADF.sample_df().show_distribution("number_range", show_percentage=False)
    t.da_export(
        Path(__file__).parent.joinpath("a.xlsx").resolve().__str__(),
        cols_contain_number=["number_range"],
        cols_contain_percentage=["percentage"],
    )
    print(t)



"""
Absfuyu: Configuration
----------------------
Package configuration module

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "ABSFUYU_CONFIG",
    "Config",
    "CONFIG_PATH",
    # "Setting"
]


# Library
# ---------------------------------------------------------------------------
from importlib.resources import files
from pathlib import Path
from typing import Any, TypedDict

from absfuyu.core import BaseClass
from absfuyu.util.json_method import JsonFile

# Setting
# ---------------------------------------------------------------------------
CONFIG_PATH = files("absfuyu.config").joinpath("config.json")
_SPACE_REPLACE = "-"  # Replace " " character in setting name


# Type hint
# ---------------------------------------------------------------------------
class SettingDictFormat(TypedDict):
    """
    Format for the ``setting`` section in ``config``

    :param default: Default value for the setting
    :param help: Description for the setting
    :param value: Current value of the setting
    """

    default: Any
    help: str
    value: Any


class ConfigFormat(TypedDict):
    """
    Config file format

    :param setting: setting section
    :type setting: dict[str, SettingDictFormat]
    :param version: version section
    :type version: VersionDictFormat
    """

    setting: dict[str, SettingDictFormat]


# Class
# ---------------------------------------------------------------------------
class Setting(BaseClass):
    """Setting"""

    def __init__(self, name: str, value: Any, default: Any, help_: str = "") -> None:
        """
        :param name: Name of the setting
        :param value: Value of the setting
        :param default: Default value of the setting
        :param help: Description of the setting
        """
        self.name = name
        self.value = value
        self.default = default
        self.help = help_

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name}: {self.value})"

    @classmethod
    def from_dict(cls, dict_data: dict[str, SettingDictFormat]):
        """
        Convert ``dict`` into ``Setting`` (``len==1`` only)
        """
        name: str = list(dict_data.keys())[0]
        _val: SettingDictFormat = list(dict_data.values())[0]
        value: Any = _val["value"]
        default: Any = _val["default"]
        help_: str = _val["help"]
        return cls(name, value, default, help_)

    def reset(self) -> None:
        """
        Reset setting to default value
        (``Setting.value = Setting.default``)
        """
        self.value = self.default

    def update_value(self, value: Any) -> None:
        """Update current value"""
        self.value = value

    def to_dict(self) -> dict[str, SettingDictFormat]:
        """
        Convert ``Setting`` into ``dict``
        """
        output: dict[str, SettingDictFormat] = {
            self.name: {"default": self.default, "help": self.help, "value": self.value}
        }
        return output


class Config(BaseClass):
    """
    Config handling
    """

    def __init__(self, config_file: Path, name: str | None = None) -> None:
        """
        config_file: Path to `.json` config file
        """
        self.config_path: Path = config_file
        self.json_engine: JsonFile = JsonFile(self.config_path)

        if name is None:
            self.name = self.config_path.name
        else:
            self.name = name

        # Data
        self.settings: list[Setting] = None  # type: ignore
        self._fetch_data()  # Load data

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.config_path.name})"

    # Data prepare and export
    def _fetch_data(self) -> None:
        """Load data from ``self.config_file`` file"""
        data: dict = self.json_engine.load_json()
        settings: dict[str, SettingDictFormat] = data.get("setting")  # type: ignore
        self.settings = [Setting.from_dict({k: v}) for k, v in settings.items()]

    def _prepare_data(self) -> ConfigFormat:
        """Prepare data to save config"""
        settings = dict()
        for setting in self.settings:
            settings.update(setting.to_dict())

        out: ConfigFormat = {"setting": settings}
        return out

    def save(self) -> None:
        """Save config to ``.json`` file"""
        self.json_engine.update_data(self._prepare_data())  # type: ignore
        self.json_engine.save_json()

    # Setting method
    @property
    def setting_list(self) -> list[str]:
        """List of name of available settings"""
        return [setting.name for setting in self.settings]

    def _get_setting(self, name: str):
        """Get setting"""
        name = name.strip().lower().replace(" ", _SPACE_REPLACE)
        if name in self.setting_list:
            for setting in self.settings:
                if setting.name.startswith(name):
                    return setting
        else:
            raise ValueError(f"Setting list: {self.setting_list}")

    def reset_config(self) -> None:
        """Reset all settings to default value"""
        [setting.reset() for setting in self.settings]  # type: ignore
        self.save()

    def show_settings(self) -> list[Setting]:
        """
        Returns a list of available settings
        (wrapper for ``Config.settings``)
        """
        return self.settings

    def change_setting(self, name: str, value: Any) -> None:
        """
        Change ``Setting`` (if available)

        Parameters
        ----------
        name : str
            Name of the setting

        value : Any
            Value of the setting
        """
        name = name.strip().lower().replace(" ", _SPACE_REPLACE)
        if name in self.setting_list:
            for setting in self.settings:
                if setting.name.startswith(name):
                    setting.update_value(value)
                    break
        else:
            raise ValueError(f"Setting list: {self.setting_list}")
        self.save()

    def toggle_setting(self, name: str) -> None:
        """
        Special ``change_setting()`` method.
        Turn on/off if ``type(<setting>) is bool``

        Parameters
        ----------
        name : str
            Name of the setting
        """
        # Get setting
        setting = self._get_setting(name)
        setting_value: bool = setting.value

        # Change value
        try:
            self.change_setting(name, not setting_value)
        except Exception:
            raise SystemExit("This setting is not type: bool")  # noqa: B904

    def add_setting(self, name: str, value: Any, default: Any, help_: str = "") -> None:
        """
        Add ``Setting`` if not exist

        Parameters
        ----------
        name : str
            Name of the setting

        value : Any
            Value of the setting

        default : Any
            Default value of the setting

        help_ : str
            Description of the setting (Default: ``None``)
        """
        name = name.strip().lower().replace(" ", _SPACE_REPLACE)
        new_setting = Setting(name, value, default, help_)
        if new_setting not in self.settings:
            self.settings.append(new_setting)
        self.save()

    def del_setting(self, name: str) -> None:
        """
        Delete ``Setting``

        Parameters
        ----------
        name : str
            Name of the setting
        """
        name = name.strip().lower().replace(" ", _SPACE_REPLACE)
        self.settings = [x for x in self.settings if x.name != name]
        self.save()

    def welcome(self) -> None:
        """Run first-run script (if any)"""
        self.change_setting("first-run", False)


# Init
# ---------------------------------------------------------------------------
ABSFUYU_CONFIG = Config(CONFIG_PATH)  # type: ignore
